import numpy as np

from picai_eval import Metrics


def test_lesion_tpr_at_fpr():
    """
    Test the lesion TPR at FPR function
    """
    lesion_results = {
        "0": [(0, 1, 1.)],
        "1": [(0, 2, 1.)],
        "2": [(1, 3, 1.)],
        "3": [(1, 1, 1.), (1, 1, 1.)],
        "4": [(0, 3, 1.)],
        "5": [(1, 2.5, 1.), (1, 1.5, 1.)],
    }
    metrics = Metrics(lesion_results=lesion_results)
    np.testing.assert_allclose(metrics.lesion_TPR, [0.2, 0.4, 0.4, 0.6, 0.6])
    np.testing.assert_allclose(metrics.lesion_FPR, [0.16666667, 0.16666667, 0.33333334, 0.33333334, np.inf])

    # test with FPR = 0.3
    np.testing.assert_almost_equal(metrics.lesion_TPR_at_FPR(0.3), 0.4)

    # test with too low FPR
    np.testing.assert_almost_equal(metrics.lesion_TPR_at_FPR(0.1), 0.0)


if __name__ == "__main__":
    test_lesion_tpr_at_fpr()



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division, print_function

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Hashable, List, Optional, Tuple, Union

import numpy as np
import sklearn
from packaging import version
from sklearn.metrics import auc, precision_recall_curve, roc_curve

try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass

from picai_eval.data_utils import PathLike, load_metrics, save_metrics


@dataclass
class Metrics:
    lesion_results: Union[Dict[Hashable, List[Tuple[int, float, float]]], PathLike]
    case_target: Optional[Dict[Hashable, int]] = None
    case_pred: Optional[Dict[Hashable, float]] = None
    case_weight: Optional[Union[Dict[Hashable, float], List[float]]] = None
    lesion_weight: Optional[Dict[Hashable, List[float]]] = None
    thresholds: "Optional[npt.NDArray[np.float64]]" = None
    subject_list: Optional[List[str]] = None
    sort: bool = True

    def __post_init__(self):
        if isinstance(self.lesion_results, (str, Path)):
            # load metrics from file
            self.load(self.lesion_results)

        if self.subject_list is None:
            self.subject_list = sorted(list(self.lesion_results))

        if self.case_target is None:
            # derive case-level targets as the maximum lesion-level target
            self.case_target = {
                idx: max([is_lesion for is_lesion, _, _ in case_y_list]) if len(case_y_list) else 0
                for idx, case_y_list in self.lesion_results.items()
            }

        if self.case_pred is None:
            # derive case-level predictions as the maximum lesion-level prediction
            self.case_pred = {
                idx: max([confidence for _, confidence, _ in case_y_list]) if len(case_y_list) else 0
                for idx, case_y_list in self.lesion_results.items()
            }

        if not isinstance(self.case_weight, dict):
            subject_list = list(self.case_target)
            if self.case_weight is None:
                self.case_weight = {idx: 1 for idx in subject_list}
            else:
                self.case_weight = {idx: weight for idx, weight in zip(subject_list, self.case_weight)}

        if self.lesion_weight is None:
            subject_list = sorted(list(self.lesion_results))
            self.lesion_weight = {idx: [1]*len(case_y_list) for idx, case_y_list in self.lesion_results.items()}

        if self.sort:
            # sort dictionaries
            subject_list = sorted(list(self.lesion_results))
            self.lesion_results = {idx: self.lesion_results[idx] for idx in subject_list}
            self.lesion_weight = {idx: self.lesion_weight[idx] for idx in subject_list}
            self.case_target = {idx: self.case_target[idx] for idx in subject_list}
            self.case_pred = {idx: self.case_pred[idx] for idx in subject_list}
            self.case_weight = {idx: self.case_weight[idx] for idx in subject_list}

    # aggregates
    def calc_auroc(self, subject_list: Optional[List[str]] = None) -> float:
        """Calculate case-level Area Under the Receiver Operating Characteristic curve (AUROC)"""
        return self.calculate_ROC(subject_list=subject_list)['AUROC']

    @property
    def auroc(self) -> float:
        """Calculate case-level Area Under the Receiver Operating Characteristic curve (AUROC)"""
        return self.calc_auroc()

    def calc_AP(self, subject_list: Optional[List[str]] = None) -> float:
        """Calculate Average Precision"""
        return self.calculate_precision_recall(subject_list=subject_list)['AP']

    @property
    def AP(self) -> float:
        """Calculate Average Precision"""
        return self.calc_AP()

    @property
    def num_cases(self) -> int:
        """Calculate the number of cases"""
        return len(self.subject_list)

    @property
    def num_lesions(self) -> int:
        """Calculate the number of ground truth lesions"""
        return sum([is_lesion for is_lesion, *_ in self.lesion_results_flat])

    @property
    def score(self):
        """Calculate the ranking score, as used in the PI-CAI 22 Grand Challenge"""
        return (self.auroc + self.AP) / 2

    def lesion_TPR_at_FPR(self, FPR: float) -> float:
        """Calculate the lesion-level true positive rate (sensitivity) at a given
        false positive rate (average number of false positives per examimation)"""
        if np.min(self.lesion_FPR) > FPR:
            return 0
        return self.lesion_TPR[self.lesion_FPR <= FPR][-1]

    # lesion-level results
    def get_lesion_results_flat(self, subject_list: Optional[List[str]] = None):
        """Flatten the per-case lesion evaluation results into a single list"""
        if subject_list is None:
            subject_list = self.subject_list

        return [
            (is_lesion, confidence, overlap)
            for subject_id in subject_list
            for is_lesion, confidence, overlap in self.lesion_results[subject_id]
        ]

    @property
    def lesion_results_flat(self) -> List[Tuple[int, float, float]]:
        """Flatten the per-case y_list"""
        return self.get_lesion_results_flat()

    def get_lesion_weight_flat(self, subject_list: Optional[List[str]] = None) -> List[float]:
        """Retrieve lesion-wise sample weights (for a given subset of cases)"""
        if subject_list is None:
            subject_list = self.subject_list

        # collect lesion weights (and flatten)
        return [weight for subject_id in subject_list for weight in self.lesion_weight[subject_id]]

    @property
    def lesion_weight_flat(self) -> List[float]:
        """Retrieve lesion-wise sample weights (for a given subset of cases)"""
        return self.get_lesion_weight_flat()

    @property
    def precision(self) -> "npt.NDArray[np.float64]":
        """Calculate lesion-level precision at each threshold"""
        return self.calculate_precision_recall()['precision']

    @property
    def recall(self) -> "npt.NDArray[np.float64]":
        """Calculate lesion-level recall at each threshold"""
        return self.calculate_precision_recall()['recall']

    @property
    def lesion_TP(self) -> "npt.NDArray[np.float64]":
        """Calculate number of true positive lesion detections at each threshold"""
        return self.calculate_counts()['TP']

    @property
    def lesion_FP(self) -> "npt.NDArray[np.float64]":
        """Calculate number of false positive lesion detections at each threshold"""
        return self.calculate_counts()['FP']

    @property
    def lesion_TPR(self) -> "npt.NDArray[np.float64]":
        """Calculate lesion-level true positive rate (sensitivity) at each threshold"""
        if self.num_lesions > 0:
            return self.lesion_TP / self.num_lesions
        else:
            return np.array([np.nan] * len(self.lesion_TP))

    @property
    def lesion_FPR(self) -> "npt.NDArray[np.float64]":
        """Calculate lesion-level false positive rate (number of false positives per case) at each threshold"""
        return self.lesion_FP / self.num_cases

    # case-level results
    def calc_case_TPR(self, subject_list: Optional[List[str]] = None) -> "npt.NDArray[np.float64]":
        """Calculate case-level true positive rate (sensitivity) at each threshold"""
        return self.calculate_ROC(subject_list=subject_list)['TPR']

    @property
    def case_TPR(self) -> "npt.NDArray[np.float64]":
        """Calculate case-level true positive rate (sensitivity) at each threshold"""
        return self.calc_case_TPR()

    def calc_case_FPR(self, subject_list: Optional[List[str]] = None) -> "npt.NDArray[np.float64]":
        """Calculate case-level false positive rate (1 - specificity) at each threshold"""
        return self.calculate_ROC(subject_list=subject_list)['FPR']

    @property
    def case_FPR(self) -> "npt.NDArray[np.float64]":
        """Calculate case-level false positive rate (1 - specificity) at each threshold"""
        return self.calc_case_FPR()

    # supporting functions
    def calculate_counts(self, subject_list: Optional[List[str]] = None) -> "Dict[str, npt.NDArray[np.float32]]":
        """
        Calculate lesion-level true positive (TP) detections and false positive (FP) detections as each threshold.
        """
        # flatten y_list (and select cases in subject_list)
        lesion_y_list = self.get_lesion_results_flat(subject_list=subject_list)

        # collect targets and predictions
        y_true: "npt.NDArray[np.float64]" = np.array([target for target, *_ in lesion_y_list])
        y_pred: "npt.NDArray[np.float64]" = np.array([pred for _, pred, *_ in lesion_y_list])

        if self.thresholds is None:
            # collect thresholds for lesion-based analysis
            self.thresholds = np.unique(y_pred)
            self.thresholds[::-1].sort()  # sort thresholds in descending order (inplace)

            # for >10,000 thresholds: resample to 10,000 unique thresholds, while also
            # keeping all thresholds higher than 0.8 and the first 20 thresholds
            if len(self.thresholds) > 10_000:
                rng = np.arange(1, len(self.thresholds), len(self.thresholds)/10_000, dtype=np.int32)
                st = [self.thresholds[i] for i in rng]
                low_thresholds = self.thresholds[-20:]
                self.thresholds = np.array([t for t in self.thresholds if t > 0.8 or t in st or t in low_thresholds])

        # define placeholders
        FP: "npt.NDArray[np.float32]" = np.zeros_like(self.thresholds, dtype=np.float32)
        TP: "npt.NDArray[np.float32]" = np.zeros_like(self.thresholds, dtype=np.float32)

        # for each threshold: count FPs and TPs
        for i, th in enumerate(self.thresholds):
            y_pred_thresholded = (y_pred >= th).astype(int)
            tp = np.sum(y_true*y_pred_thresholded)
            fp = np.sum(y_pred_thresholded - y_true*y_pred_thresholded)

            # update with new point
            FP[i] = fp
            TP[i] = tp

        # extend curve to infinity
        TP[-1] = TP[-2]
        FP[-1] = np.inf

        return {
            'TP': TP,
            'FP': FP,
        }

    def calculate_precision_recall(self, subject_list: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Generate Precision-Recall curve and calculate average precision (AP).
        """
        # flatten y_list (and select cases in subject_list)
        lesion_y_list = self.get_lesion_results_flat(subject_list=subject_list)

        # collect targets and predictions
        y_true: "npt.NDArray[np.float64]" = np.array([target for target, *_ in lesion_y_list])
        y_pred: "npt.NDArray[np.float64]" = np.array([pred for _, pred, *_ in lesion_y_list])

        # calculate precision-recall curve
        if version.parse(sklearn.__version__) >= version.parse("1.5"):
            # in the future this if/else block can be removed, then set 1.5 as minimum in requirements.txt
            precision, recall, thresholds = precision_recall_curve(
                y_true=y_true,
                y_score=y_pred,
                sample_weight=self.get_lesion_weight_flat(subject_list=subject_list)
            )
        else:
            precision, recall, thresholds = precision_recall_curve(
                y_true=y_true,
                probas_pred=y_pred,
                sample_weight=self.get_lesion_weight_flat(subject_list=subject_list)
            )

        # set precision to zero at a threshold of "zero", as those lesion
        # candidates are included just to convey the number of lesions to
        # the precision_recall_curve function, and not actual candidates
        precision[:-1][thresholds == 0] = 0

        # calculate average precision using the step function integral
        # The following works because the last entry of precision is
        # guaranteed to be 1, as returned by precision_recall_curve
        # Taken from https://github.com/scikit-learn/scikit-learn/blob/
        # 32f9deaaf27c7ae56898222be9d820ba0fd1054f/sklearn/metrics/_ranking.py#L212
        AP = -np.sum(np.diff(recall) * np.array(precision)[:-1])

        return {
            'AP': AP,
            'precision': precision,
            'recall': recall,
        }

    def calculate_ROC(self, subject_list: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Generate Receiver Operating Characteristic curve for case-level risk stratification.
        """
        if subject_list is None:
            subject_list = self.subject_list

        fpr, tpr, _ = roc_curve(
            y_true=[self.case_target[s] for s in subject_list],
            y_score=[self.case_pred[s] for s in subject_list],
            sample_weight=[self.case_weight[s] for s in subject_list],
        )

        auroc = auc(fpr, tpr)

        return {
            'FPR': fpr,
            'TPR': tpr,
            'AUROC': auroc,
        }

    @property
    def version(self):
        return "1.4.x"

    def as_dict(self):
        return {
            # aggregates
            "auroc": self.auroc,
            "AP": self.AP,
            "num_cases": self.num_cases,
            "num_lesions": self.num_lesions,
            "picai_eval_version": self.version,

            # lesion-level results
            "lesion_results": self.lesion_results,
            "lesion_weight": self.lesion_weight,

            # case-level results
            "case_pred": self.case_pred,
            "case_target": self.case_target,
            "case_weight": self.case_weight,
        }

    def full_dict(self):
        return {
            # aggregates
            "auroc": self.auroc,
            "AP": self.AP,
            "num_cases": self.num_cases,
            "num_lesions": self.num_lesions,
            "picai_eval_version": self.version,

            # lesion-level results
            "lesion_results": self.lesion_results,
            "lesion_weight": self.lesion_weight,
            "precision": self.precision,
            "recall": self.recall,
            'lesion_TPR': self.lesion_TPR,
            'lesion_FPR': self.lesion_FPR,
            "thresholds": self.thresholds,

            # case-level results
            "case_pred": self.case_pred,
            "case_target": self.case_target,
            "case_weight": self.case_weight,
        }

    def minimal_dict(self):
        return {
            # lesion-level results
            "lesion_results": self.lesion_results,
            "lesion_weight": self.lesion_weight,

            # case-level results
            "case_pred": self.case_pred,
            "case_target": self.case_target,
            "case_weight": self.case_weight,
        }

    def save(self, path: PathLike):
        """Save metrics to file (including aggregates)"""
        save_metrics(metrics=self.as_dict(), file_path=path)

    def save_full(self, path: PathLike):
        """Save metrics to file (including derived metrics)"""
        save_metrics(metrics=self.full_dict(), file_path=path)

    def save_minimal(self, path: PathLike):
        """Save metrics to file (minimal required metrics)"""
        save_metrics(metrics=self.minimal_dict(), file_path=path)

    def load(self, path: PathLike):
        """Load metrics from file"""
        metrics = load_metrics(path)

        # parse metrics
        self.case_target = {idx: int(float(val)) for idx, val in metrics['case_target'].items()}
        self.case_pred = {idx: float(val) for idx, val in metrics['case_pred'].items()}
        self.case_weight = {idx: float(val) for idx, val in metrics['case_weight'].items()}
        self.lesion_weight = {idx: [float(val) for val in weights] for idx, weights in metrics['lesion_weight'].items()}
        self.lesion_results = {
            idx: [
                (int(float(is_lesion)), float(confidence), float(overlap))
                for (is_lesion, confidence, overlap) in lesion_results_case
            ]
            for idx, lesion_results_case in metrics['lesion_results'].items()
        }

    def __str__(self) -> str:
        return f"Metrics(auroc={self.auroc:.2%}, AP={self.AP:.2%}, {self.num_cases} cases, {self.num_lesions} lesions)"

    def __repr__(self) -> str:
        return self.__str__()



import os

import pytest
from numpy.testing import assert_allclose

from picai_eval.eval import evaluate_case


@pytest.mark.parametrize("y_det,y_true,expected_y_list", [
    # Case 0: no ground truth lesions, no detection maps.
    ("case-0_detection_map.nii.gz", "case-0_label.nii.gz", []),

    # Case 1: single ground truth lesion, no detection maps. 1 FN.
    ("case-1_detection_map.nii.gz", "case-1_label.nii.gz", [(1, 0.0, 0.)]),

    # Case 2: no ground truth lesion, single ground truth lesion. 1 FP.
    ("case-2_detection_map.nii.gz", "case-2_label.nii.gz", [(0, 1.0, 0.)]),

    # Case 3: single ground truth lesion, single detection map. Candidate 1: IoU = 0. 1 FN + 1 FP.
    ("case-3_detection_map.nii.gz", "case-3_label.nii.gz", [(1, 0.0, 0.), (0, 1.0, 0.0)]),

    # Case 4: single ground truth lesion, single detection map. Candidate 1: IoU = 1/3. 1 TP.
    ("case-4_detection_map.nii.gz", "case-4_label.nii.gz", [(1, 1.0, 1/3)]),

    # Case 5: single ground truth lesion, two detection maps. Candidate 1: IoU = 9/41, confidence = 1. Candidate 2: IoU = 3/47, confidence = 2. 1 TP + 1 FP.
    ("case-5_detection_map.nii.gz", "case-5_label.nii.gz", [(1, 1.0, 9/41), (0, 2.0, 0)]),

    # Case 6: single ground truth lesion, two detection maps. Candidate 1: IoU = 1/3, confidence = 1. Candidate 2: IoU = 2/7, confidence = 2. 1 TP.
    ("case-6_detection_map.nii.gz", "case-6_label.nii.gz", [(1, 1.0, 1/3)]),

    # Case 7: two ground truth lesion, single detection map. Candidate 1: IoU = 1/3 with either ground truth lesion. 1 TP + 1 FN.
    ("case-7_detection_map.nii.gz", "case-7_label.nii.gz", [(1, 1.0, 1/3), (1, 0.0, 0.0)]),

    # Case 8: two ground truth lesion, single detection map. Candidate 1: IoU = 5/11 with one gt lesion, 1/11 with other gt lesion. 1 TP + 1 FN.
    ("case-8_detection_map.nii.gz", "case-8_label.nii.gz", [(1, 1.0, 5/11), (1, 0.0, 0.0)]),

    # Case 9: single ground truth lesion, one lesion detection map. Candidate 1: IoU = 1/25. 1 FN + 1 FP.
    ("case-9_detection_map.nii.gz", "case-9_label.nii.gz", [(1, 0.0, 0.0), (0, 1.0, 0.0)]),
])
def test_evaluate_case(y_det, y_true, expected_y_list):
    """
    Lesion candidate extraction and FROC evaluation Unit Test
    Evaluate predictions and calculate FROC statistics
    """

    y_list, *_ = evaluate_case(
        y_det=os.path.join("tests/test-maps", y_det),
        y_true=os.path.join("tests/test-maps", y_true),
    )

    assert_allclose(y_list, expected_y_list)



from typing import List

import numpy as np
import pytest
from numpy.testing import assert_allclose

from picai_eval.statistical_helper import (calc_sensitivity, calc_specificity,
                                           match_reader, match_then_compare,
                                           multiple_permutation_tests,
                                           perform_matched_boostrapping)


def sample_predictions(n_pos=300, n_neg=700, kwargs_pos=None, kwargs_neg=None, seed=None):
    """Generate a set of predictions. With default parameters, these predictions have ~0.9 AUROC."""
    # set default values (which have ~0.9 AUC)
    if kwargs_pos is None:
        kwargs_pos = dict(a=3, b=2)
    if kwargs_neg is None:
        kwargs_neg = dict(a=2, b=6)

    # set random seed
    np.random.seed(seed)

    # sample predictions
    y_pred = np.concatenate((
        np.random.beta(size=n_pos, **kwargs_pos),
        np.random.beta(size=n_neg, **kwargs_neg),
    ))

    return y_pred


def sample_prediction_set(instances=10, n_pos=300, n_neg=700, kwargs_pos=None, kwargs_neg=None, seed=None):
    """Generate multiple sets of predictions and a set of labels."""
    # set random seed
    np.random.seed(seed)
    seeds = np.random.randint(int(1e5), size=instances)

    # create labels
    y_true = np.array([1] * n_pos + [0] * n_neg)

    y_pred = np.array([
        sample_predictions(
            n_pos=n_pos,
            n_neg=n_neg,
            kwargs_pos=kwargs_pos,
            kwargs_neg=kwargs_neg,
            seed=seed
        )
        for seed in seeds
    ])

    return y_true, y_pred


def test_multiple_permutation_tests_exact():
    # perform multiple permutation tests
    p_values = multiple_permutation_tests(
        all_scores={
            'a': {
                'scores_baseline': [0, 1, 2],
                'scores_alternative': [1, 2, 3]},
            'b': {
                'scores_baseline': [4, 1, 2],
                'scores_alternative': [1, 2, 3]
            },
            'c': {
                'scores_baseline': [0.92, 0.94, 0.95, 0.81, 0.82, 0.86],
                'scores_alternative': [0.96, 0.91, 0.90, 0.85, 0.81, 0.80]
            },
        }
    )

    # check p-values (are calculated exactly)
    for key in p_values:
        assert_allclose(p_values[key], {'a': 0.25, 'b': 0.7, 'c': 0.7218614718614719}[key])


# note: not all predictions set can be exactly matched, so always check for bias!
# the seed below, 1, generates predictions that can be matched.
@pytest.mark.parametrize("match", [
    'sensitivity',
    'specificity',
])
def test_operating_point_matching(match, iterations=1000, seed=576):
    # sample a prediction set, take one set of predictions as reader and another as AI predictions
    y_true, y_pred = sample_prediction_set(seed=seed)
    y_pred_reader = (y_pred[0] > 0.5).astype(int)
    y_pred_ai = y_pred[1]

    # sample indices, match performance and check performance difference (to discover bias in matching)
    np.random.seed(seed)
    z: List[float] = []
    for _ in range(iterations):
        indices = np.random.randint(0, len(y_true), len(y_true))
        if len(np.unique(y_true[indices])) < 2:
            continue

        y_true_it = y_true[indices]
        y_pred_ai_it = y_pred_ai[indices]
        y_pred_reader_it = y_pred_reader[indices]

        # match AI operating point to the reader
        y_pred_ai_matched_it = match_reader(
            y_true=y_true_it,
            y_pred_ai=y_pred_ai_it,
            y_pred_reader=y_pred_reader_it,
            match=match
        )

        # check performance difference between reader and matched AI
        if match == "sensitivity":
            performance_reader = calc_sensitivity(y_true=y_true_it, y_pred=y_pred_reader_it)
            performance_ai = calc_sensitivity(y_true=y_true_it, y_pred=y_pred_ai_matched_it)
        elif match == 'specificity':
            performance_reader = calc_specificity(y_true=y_true_it, y_pred=y_pred_reader_it)
            performance_ai = calc_specificity(y_true=y_true_it, y_pred=y_pred_ai_matched_it)
        else:
            raise NotImplementedError

        # collect performance difference
        z.append(performance_reader - performance_ai)

    # check if performance difference is zero for all subsets
    assert_allclose(np.mean(z), 0, atol=1e-3)


# The expected differences in Cohen's κ listed below are not based on manual calculations.
# Given that the two sets of predictions are sampled from the same distribution,
# the κ's should be close to each other (and therefore a difference close to zero).
@pytest.mark.parametrize("match,expected_z", [
    ('sensitivity', -0.025217179012471225),
    ('specificity', -0.025721235660373654),
])
def test_match_then_compare(match, expected_z, seed=576):
    # sample a prediction set, take one set of predictions as reader and another as AI predictions
    y_true, y_pred = sample_prediction_set(seed=seed)
    y_pred_reader = (y_pred[0] > 0.5).astype(int)
    y_pred_ai = y_pred[1]

    # perform matching and calculate difference in Cohen's κ
    z = match_then_compare(
        y_true=y_true,
        y_pred_ai=y_pred_ai,
        y_pred_reader=y_pred_reader,
        match=match,
    )

    # check if Cohen's κ difference (reader κ - AI κ) is as expected
    print("Observed z:", z)
    assert_allclose(z, expected_z)


# note: not all predictions set can be exactly matched, so always check for bias!
# the seed below, 1, generates predictions that can be matched.
def test_bootstrapping_sample_weights(match='sensitivity', flavour='match>sample>compare', seed=1):
    # sample a prediction set, take one set of predictions as reader and the others as AI predictions
    y_true, y_pred = sample_prediction_set(seed=seed)
    y_pred_reader = (y_pred[0] > 0.5).astype(int)
    y_pred_ai = y_pred[1:]

    kwargs = dict(
        y_true=y_true,
        y_pred_ai=y_pred_ai,
        y_pred_reader=y_pred_reader,
        match=match,
        flavour=flavour,
        iterations=100,
        seed=seed
    )

    # perform matched bootstrapping (twice)
    perform_matched_boostrapping(**kwargs)  # warmup. Should not be required, but it is.
    p1 = perform_matched_boostrapping(**kwargs)
    p2 = perform_matched_boostrapping(**kwargs)

    # perform matched bootstrapping with dummy sample weights
    p3 = perform_matched_boostrapping(
        sample_weight=np.random.uniform(size=len(y_true)),
        **kwargs
    )

    # compare p-values
    assert p1 == p2, "Random seed was not respected"
    assert p1 != p3, "Sample weight was not respected"



import json

import numpy as np
from numpy.testing import assert_allclose
from picai_eval.analysis_utils import (calculate_pAUC_from_graph,
                                       calculate_statistics)


def test_calculate_statistics():
    """
    Unit Test: Calculating operating points
    """

    with open("tests/test-metrics.json") as fp:
        # read metrics with the following structure:
        # {
        #   'tpr': [[y1m1, y2m1, ...], [y1m2, y2m2, ...], ...]
        #   'fpr': [[x1m1, x2m1, ...], [x1m2, x2m2, ...], ...]
        # }
        metrics = json.load(fp)

    # restructure metrics
    metrics = list(zip(metrics['tpr'], metrics['fpr']))

    # calculate statistics
    op_match = {'y': 0.9166666666666666, 'x': 0.2205882353}  # x: 1-0.7794117647058824
    results = calculate_statistics(metrics=metrics, op_match=op_match)

    # Unit tests:
    assert_allclose(results['auroc_mean'], 0.8976800789760347), "Test 1 (AUROC) for calculating statistics failed!"
    assert_allclose(results['op_closest_y_y'], 0.9180555555555557), "Test 2 (operating point, sensitivity) for calculating statistics failed!"
    assert_allclose(results['op_closest_y_x'], 0.35294117647058826), "Test 3 (operating point, specificity) for calculating statistics failed!"  # 1 - 0.647
    assert_allclose(results['op_closest_y_y_std'], 0.0025983731852596724), "Test 4 (operating point, sensitivity std.) for calculating statistics failed!"
    assert_allclose(results['op_closest_y_x_std'], 0.06562039423796671), "Test 5 (operating point, specificity std.) for calculating statistics failed!"


def test_calculate_pAUC_from_graph():
    """
    Test normal operation for calculating the area under a curve
    """
    # test calculation of area under the curve
    # note: a typical ROC/FROC curve consists only of straight horizontal/vertical lines
    y = [0.00, 0.13, 0.13, 0.15, 0.15, 0.17, 0.17, 0.17, 0.50, 0.50, 0.89, 0.89, 1.00, 1.00]
    x = [0.00, 0.00, 0.10, 0.10, 0.20, 0.20, 0.30, 0.40, 0.40, 0.50, 0.50, 0.60, 0.60, 1.00]
    pAUC = calculate_pAUC_from_graph(
        x=x, y=y, pAUC_start=0.0, pAUC_end=1.0,
    )[0]
    assert_allclose(pAUC, (0.13 + 0.15 + 2*0.17 + 0.50 + 0.89 + 4*1.00) / 10)

    # test calculation of partial area under the curve
    y = [0.00, 0.13, 0.13, 0.15, 0.15, 0.17, 0.17, 0.17, 0.50, 0.50, 0.89, 0.89, 1.00, 1.00]
    x = [0.00, 0.00, 0.10, 0.10, 0.20, 0.20, 0.30, 0.40, 0.40, 0.50, 0.50, 0.60, 0.60, 1.00]
    pAUC = calculate_pAUC_from_graph(
        x=x, y=y, pAUC_start=0.20, pAUC_end=1.0,
    )[0]
    assert_allclose(pAUC, (2*0.17 + 0.50 + 0.89 + 4*1.00) / 10)

    # test partial AUC
    y = np.linspace(0, 1, num=101)
    x = np.linspace(0, 1, num=101)  # <- need value at 0.1 for pAUC to be exactly as calculated
    pAUC = calculate_pAUC_from_graph(
        x=x, y=y, pAUC_start=0.1, pAUC_end=1.0,
    )[0]
    assert_allclose(pAUC, 1/2 - 0.1*0.1/2)


def test_calculate_pAUC_from_graph_edge_cases():
    """
    Test edge cases for calculating the area under a curve
    """
    # ensure a warning is displayed when calculating statistics for a non-monotonically increasing curve
    # note: if the warning messages are not displayed, the coverage should not be 100%!
    y = [0, 1, 2, 3, 4, 3, 2, 3, 4, 5]
    x = [0, 1, 5, 6, 4, 3, 2, 3, 4, 5]
    calculate_pAUC_from_graph(
        x=x, y=y, pAUC_start=0.1, pAUC_end=3.1415,
    )

    # test unpacking of x and y
    y = [[np.linspace(0, 1)]]
    x = [[np.linspace(0, 1)]]
    pAUC = calculate_pAUC_from_graph(
        x=x, y=y, pAUC_start=0., pAUC_end=1.0,
    )[0]
    assert_allclose(pAUC, 1/2)

    # test return full = True
    y = np.linspace(0, 1)
    x = np.linspace(0, 1)
    results = calculate_pAUC_from_graph(
        x=x, y=y, pAUC_start=0., pAUC_end=1.0, full=True,
    )
    assert_allclose(results['auroc_all'], [1/2])



import os
import sys
from subprocess import check_call

import numpy as np
import pytest
from numpy.testing import assert_allclose
from sklearn.metrics import auc, roc_curve

from picai_eval import evaluate
from picai_eval.data_utils import load_metrics
from picai_eval.eval import Metrics, evaluate_folder
from picai_eval.image_utils import (read_label, read_prediction,
                                    resize_image_with_crop_or_pad)

subject_list = [
    f"case-{i}"
    for i in range(10)
]
lesion_results_expected = [
    (1, 0.0, 0.),
    (0, 1.0, 0.),
    (1, 0.0, 0.), (0, 1.0, 0.0),
    (1, 1.0, 1/3),
    (1, 1.0, 9/41), (0, 2.0, 0),
    (1, 1.0, 1/3),
    (1, 1.0, 1/3), (1, 0.0, 0.0),
    (1, 1.0, 5/11), (1, 0.0, 0.0),
    (1, 0.0, 0.0), (0, 1.0, 0.0),
]
in_dir_det = "tests/test-maps/"
in_dir_annot = "tests/test-maps/"


@pytest.fixture(scope="session")
def y_det():
    """
    Read crafted detection maps
    """
    assert os.path.exists(in_dir_det), "Folder with softmax files not found!"
    y_det = [
        read_prediction(os.path.join(in_dir_det, f"{subject_id}_detection_map.nii.gz"))
        for subject_id in subject_list
    ]

    yield y_det


@pytest.fixture(scope="session")
def y_true():
    """
    Read crafted labels
    """
    assert os.path.exists(in_dir_annot), "Folder with annotations not found!"
    y_true = [
        read_label(os.path.join(in_dir_annot, f"{subject_id}_label.nii.gz"))
        for subject_id in subject_list
    ]

    yield y_true


def test_evaluation(y_det, y_true, num_parallel_calls=3):
    """
    Test standard evaluation pipeline
    The 10 crafted cases in subject_list should have:
    - TPs: 5
    - FPs: 4
    - FNs: 5
    """

    metrics = evaluate(
        y_det=y_det,
        y_true=y_true,
        subject_list=subject_list,
        num_parallel_calls=num_parallel_calls,
    )

    # check metrics
    assert metrics.lesion_TP[-2] == 5
    assert metrics.lesion_FP[-2] == 4
    assert metrics.num_lesions - metrics.lesion_TP[-2] == 5
    assert metrics.AP == (5/9)*(1/2) + 0*(1/2)

    # check lesion_results
    lesion_results_flat = [
        (is_lesion, confidence, overlap)
        for subject_id in subject_list
        for is_lesion, confidence, overlap in metrics.lesion_results[subject_id]
    ]
    assert_allclose(lesion_results_expected, lesion_results_flat)


def test_evaluation_DSC(y_det, y_true):
    """
    Test DSC overlap function
    """
    metrics = evaluate(
        y_det=y_det,
        y_true=y_true,
        subject_list=subject_list,
        overlap_func="DSC"
    )

    # check metrics
    assert metrics.lesion_TP[-2] == 5
    assert metrics.lesion_FP[-2] == 4
    assert metrics.num_lesions - metrics.lesion_TP[-2] == 5
    assert metrics.AP == (5/9)*(1/2) + 0*(1/2)


def test_evaluation_prediction_padding(y_det, y_true):
    """
    Test if functions will pad detection maps
    """

    y_true_padded = [
        resize_image_with_crop_or_pad(lbl, img_size=np.array(lbl.shape)+2)
        for lbl in y_true
    ]

    assert y_true_padded[0].shape != y_det[0].shape

    metrics = evaluate(
        y_det=y_det,
        y_true=y_true_padded,
        subject_list=subject_list
    )

    # check lesion_results
    lesion_results_flat = [
        (is_lesion, confidence, overlap)
        for subject_id in subject_list
        for is_lesion, confidence, overlap in metrics.lesion_results[subject_id]
    ]
    assert_allclose(lesion_results_expected, lesion_results_flat)


def test_sample_weights():
    """
    Test if functions apply sample weights correctly
    """

    lesion_results = {
        "0": [(0, 1, 1.)],
        "1": [(0, 2, 1.)],
        "2": [(1, 3, 1.)],
        "3": [(1, 1, 1.), (1, 1, 1.)],
    }

    lesion_weight = {
        "0": [1],
        "1": [1],
        "2": [1],
        "3": [4, 5],
    }

    # calculate metrics without weights
    metrics = Metrics(lesion_results=lesion_results)

    # calculate metrics with case/leison-level weights
    metrics_weighted = Metrics(
        lesion_results=lesion_results,
        case_weight=np.random.uniform(size=metrics.num_cases),
        lesion_weight=lesion_weight
    )

    # verify weights are applied
    assert metrics.auroc != metrics_weighted.auroc, "Sample weight not applied to AUROC!"
    assert metrics.AP != metrics_weighted.AP, f"Sample weight not applied to AP! ({metrics.AP} == {metrics_weighted.AP})"
    assert metrics_weighted.AP == 1*(1/10) + (10/12)*(9/10), f"Sample weight not applied correctly to AP! ({metrics.AP} != 0.85)"

    # check if sample weights are applied as expected
    # example adapted from: https://rdrr.io/cran/WeightedROC/f/inst/doc/Definition.pdf
    fpr, tpr, _ = roc_curve(
        y_true=(0, 0, 1, 1, 1),
        y_score=(1, 2, 3, 1, 1),
        sample_weight=(1, 1, 1, 4, 5),
    )
    auc_score = auc(fpr, tpr)
    assert auc_score == 0.325
    assert_allclose(tpr, [0., 0.1, 0.1, 1.])
    assert_allclose(fpr, [0., 0.,  0.5, 1.])


@pytest.mark.xfail
def test_evaluation_negative_predictions(y_det, y_true):
    """
    Test if evaluation works if (some) confidence scores are negative
    """
    # shift confidence scores to include both positive and negative confidences
    shifted_y_det = [y_det - 1.5 for y_det in y_det]

    if np.min(shifted_y_det) > 0 or np.max(shifted_y_det) < 0:
        # This test requires a sufficiently spread out prediction set, don't
        # trigger the expected fail
        return

    evaluate(
        y_det=shifted_y_det,
        y_true=y_true,
        subject_list=subject_list,
    )


@pytest.mark.xfail
def test_softmax_input(y_det, y_true, num_parallel_calls=3):
    """
    Test if evaluation throws an error when the input is a softmax volume (instead of detection maps)
    """
    evaluate(
        y_det=[np.random.normal(size=pred.shape) for pred in y_det],
        y_true=y_true,
        subject_list=subject_list,
        num_parallel_calls=num_parallel_calls,
    )


def test_evaluation_from_dir_with_subject_list(num_parallel_calls=3):
    detection_map_dir = "tests/test-maps"
    metrics = evaluate_folder(
        detection_map_dir,
        subject_list=subject_list,
        num_parallel_calls=num_parallel_calls,
        verbose=1,
    )

    # check metrics
    assert metrics.lesion_TP[-2] == 5
    assert metrics.lesion_FP[-2] == 4
    assert metrics.num_lesions - metrics.lesion_TP[-2] == 5
    assert metrics.AP == (5/9)*(1/2) + 0*(1/2)


def test_evaluation_from_dir_without_subject_list(num_parallel_calls=3):
    detection_map_dir = "tests/test-maps"
    metrics = evaluate_folder(detection_map_dir, num_parallel_calls=num_parallel_calls, verbose=1)

    # check metrics
    assert metrics.lesion_TP[-2] == 5
    assert metrics.lesion_FP[-2] == 4
    assert metrics.num_lesions - metrics.lesion_TP[-2] == 5
    assert metrics.AP == (5/9)*(1/2) + 0*(1/2)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="GitHub Actions' Windows does not like check_call")
def test_main(metrics_path="tests/output/metrics.json"):
    """Test usage from command line"""
    if os.path.exists(metrics_path):
        os.remove(metrics_path)

    check_call([
        "python", "-m", "picai_eval",
        "--input", os.path.join(os.getcwd(), "tests/test-maps"),
        "--output", os.path.join(os.getcwd(), metrics_path)
    ])

    metrics = load_metrics(metrics_path)
    metrics_expected = load_metrics("tests/output-expected/metrics-expected.json")

    assert metrics == metrics_expected


def test_metrics_save(y_det, y_true, metrics_path="tests/output/metrics.json"):
    """Test saving of metrics"""
    if os.path.exists(metrics_path):
        os.remove(metrics_path)

    metrics = evaluate(
        y_det=y_det,
        y_true=y_true,
        subject_list=subject_list,
    )

    metrics.save(metrics_path)
    metrics_reloaded = Metrics(metrics_path)
    metrics_expected = Metrics("tests/output-expected/metrics-expected.json")

    assert metrics == metrics_reloaded == metrics_expected


def test_minimal_metrics_save(y_det, y_true, metrics_path="tests/output/metrics-minimal.json"):
    """Test saving of metrics"""
    if os.path.exists(metrics_path):
        os.remove(metrics_path)

    metrics = evaluate(
        y_det=y_det,
        y_true=y_true,
        subject_list=subject_list,
    )

    metrics.save_minimal(metrics_path)
    metrics_reloaded = load_metrics(metrics_path)
    metrics_expected = load_metrics("tests/output-expected/metrics-minimal-expected.json")

    assert metrics_reloaded == metrics_expected


def test_full_metrics_save(y_det, y_true, metrics_path="tests/output/metrics-full.json"):
    """Test saving of metrics"""
    if os.path.exists(metrics_path):
        os.remove(metrics_path)

    metrics = evaluate(
        y_det=y_det,
        y_true=y_true,
        subject_list=subject_list,
    )

    metrics.save_full(metrics_path)
    metrics_reloaded = load_metrics(metrics_path)
    metrics_expected = load_metrics("tests/output-expected/metrics-full-expected.json")

    assert metrics_reloaded == metrics_expected


def test_select_subset_lesion_results():
    lesion_results = {
        "0": [(0, 1, 1.)],
        "1": [(0, 2, 1.)],
        "2": [(1, 3, 1.)],
        "3": [(1, 1, 1.), (1, 1, 1.)],
    }
    lesion_results_subset_expected = [
        (1, 3, 1.), (1, 1, 1.), (1, 1, 1.)
    ]

    # calculate metrics without weights
    metrics = Metrics(lesion_results=lesion_results)

    lesion_results_subset = metrics.get_lesion_results_flat(subject_list=["2", "3"])

    assert lesion_results_subset_expected == lesion_results_subset


def test_single_threaded(y_det, y_true):
    """Test if single threaded evaluation works"""
    test_evaluation(y_det=y_det, y_true=y_true, num_parallel_calls=1)
    test_evaluation_from_dir_with_subject_list(num_parallel_calls=1)
    test_evaluation_from_dir_without_subject_list(num_parallel_calls=1)
